/***************************************************************************
  $RCSfile: ctcard.cpp,v $
                             -------------------
    cvs         : $Id: ctcard.cpp,v 1.26 2003/02/17 04:17:46 aquamaniac Exp $
    begin       : Fri Dec 06 2002
    copyright   : (C) 2002 by Martin Preuss
    email       : martin@libchipcard.de


 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef __declspec
# if BUILDING_CHIPCARD_DLL
#  define CHIPCARD_API __declspec (dllexport)
# else /* Not BUILDING_CHIPCARD_DLL */
#  define CHIPCARD_API __declspec (dllimport)
# endif /* Not BUILDING_CHIPCARD_DLL */
#else
# define CHIPCARD_API
#endif


#include "ctcard.h"
#include "ctmisc.h"
#include "libchipcard.h"
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#include <stdio.h>

#include <engine/chameleon/error.h>
#include <engine/chameleon/debug.h>

#define CTCARD_BOUNDARY 127
#define CTCARD_READSIZE 255


class LibChipCard_Initializer {
private:
  bool _isOpen;
public:
  LibChipCard_Initializer():_isOpen(false){
  };
  ~LibChipCard_Initializer(){
    _isOpen=false;
    ChipCard_Fini();
  }

  int initialize(const char *cfgfile,
		 const char *section) {
    int rv;

    if (_isOpen)
      return CHIPCARD_SUCCESS;
    rv=ChipCard_Init(cfgfile, section);
    if (rv!=CHIPCARD_SUCCESS)
      return rv;
    _isOpen=true;
    return CHIPCARD_SUCCESS;
  }
};


static LibChipCard_Initializer LibChipCard_Initializer_Object;


int CTCard::_openLibrary() {
  return LibChipCard_Initializer_Object.initialize(0,0);
}


CTCard::CallBackResult CTCard::doCallBack(bool first) {
  if (_callback)
    return (*_callback)(first, _callbackParam);
  else
    return CallBackContinue;
}


int CTCard::_responseLoop(int reqid, int timeout) {
  int err;
  int i;
  time_t startTime;
  CallBackResult rv;

  i=0;
  startTime=time(NULL);
  while(1) {
    if (timeout) {
      if (difftime(time(NULL),startTime)>=timeout) {
#if DEBUGMODE>1
	fprintf(stderr,"Timeout in responseLoop\n");
#endif
	return CHIPCARD_ERROR_NO_MESSAGE;
      }
    }
    err=ChipCard_CheckResponse(reqid);
    if (err!=CHIPCARD_SUCCESS) {
      if (err!=CHIPCARD_ERROR_NO_MESSAGE) {
#if DEBUGMODE>1
	fprintf(stderr,"Chipcard error %d\n",err);
#endif
	if (err==CHIPCARD_ERROR_UNREACHABLE ||
	    err==CHIPCARD_ERROR_NO_REQUEST ||
	    err==CHIPCARD_ERROR_INTERRUPTED)
	  return err;
      }
    }
    else {
      return CHIPCARD_SUCCESS;
    }
    err=ChipCard_Work();
    if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
      fprintf(stderr,"Chipcard error %d\n",err);
#endif
      if (err==CHIPCARD_ERROR_INTERRUPTED ||
	  err==CHIPCARD_ERROR_NO_TRANSPORT)
	return err;
    }
    rv=doCallBack(i==0);
    if (rv==CallBackAbort)
      return CHIPCARD_ERROR_ABORTED;
    i++;
  } /* while */

  return CHIPCARD_ERROR_NO_MESSAGE;
}


int CTCard::_findTerminal(int *termid,
			  unsigned int flags,
			  unsigned int flagsMask) {
  int err;
  int requestid;
  unsigned int readerbuffer[1];
  int readerBufferLength;

  err=_openLibrary();
  if (err!=CHIPCARD_SUCCESS)
    return err;

  /* find matching reader */

  err=ChipCard_RequestFindReader(&requestid,
				 "",
				 flags,
				 flagsMask);
  if (err!=CHIPCARD_SUCCESS)
    return err;

  err=_responseLoop(requestid, _timeout);
  if (err!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"1:Chipcard error %d\n",err);
#endif
    return 1;
  }

  readerBufferLength=sizeof(readerbuffer);
  err=ChipCard_CheckFindReader(requestid,
			       readerbuffer,
			       &readerBufferLength);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"2:Chipcard error %d\n",err);
#endif
    return 1;
  }

  *termid=readerbuffer[0];
  return 0;
}


int CTCard::_allocTerminal(int termid, int *termhandle,
			   CHIPCARD_READERDESCR **rd) {
  int err;
  int requestid;
  CHIPCARD_READERDESCR *lrd;
  int lth;

#if DEBUGMODE>1
  fprintf(stderr,"Will allocate terminal %08x\n",termid);
#endif
  err=_openLibrary();
  if (err!=CHIPCARD_SUCCESS)
    return err;

  /* allocate reader */
  err=ChipCard_RequestAllocReader(&requestid,
				  termid);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"Chipcard error %d",err);
#endif
    return 1;
  }

  err=_responseLoop(requestid, _timeout);
  if (err!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"No response\n");
#endif
    return 1;
  }
  err=ChipCard_CheckAllocReader(requestid,&lth,&lrd);
  if (err!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"No response\n");
#endif
    return 1;
  }

  *rd=lrd;
  *termhandle=lth;
  return 0;
}


int CTCard::_connectTerminal(int tid) {
  int err;
  int requestid;
  char atr[300];
  int atrlen;
  int result;
  int i1, i2;
  int j1, j2;


  /* find matching reader */
#if DEBUGMODE>1
  fprintf(stderr,"Will connect to %08x\n",tid);
#endif
  err=ChipCard_RequestConnect(&requestid,tid);
  if (err!=CHIPCARD_SUCCESS)
    return err;

  err=_responseLoop(requestid, _timeout);
  if (err!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"No response\n");
#endif
    return 1;
  }

  atrlen=sizeof(atr);
  err=ChipCard_CheckConnect(requestid,
                            &result,
			    atr,
			    &atrlen);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"Chipcard error %d\n",err);
#endif
    return err;
  }
  if (result!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"Could not connect (%d)\n", result);
#endif
    return CHIPCARD_ERROR_NO_CARD;
  }
  // set memory size
  if (atrlen>=2) {
    i1=(atr[1]>>3) & 0x07; // count of elements
    i2=atr[1] & 0x07;      // size of element

    // check element size
    if (i2==0)
      j2=0;
    else if (i2==1)
      j2=1;
    else
      j2=1<<i2;

    // check element number
    if (i1==0)
      j1=0;
    else
      j1=64<<i1;

    // calculate memory size
    if (j1 && j2)
      _memsize=j1*j2/8;
    else
      _memsize=0;
  }
  else
    _memsize=0;

  _atr=string(atr, atrlen);

  return 0;
}


int CTCard::_disconnectTerminal(int tid) {
  int err;
  int requestid;
  int result;

  /* find matching reader */
  err=ChipCard_RequestDisconnect(&requestid,tid);
  if (err!=CHIPCARD_SUCCESS)
    return err;

  err=_responseLoop(requestid, _timeout);
  if (err!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"No response.\n");
#endif
    return 1;
  }

  err=ChipCard_CheckDisconnect(requestid,
			       &result);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"Chipcard error %d\n",err);
#endif
    return 1;
  }
  if (result!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"Could not disconnect\n");
#endif
    return 1;
  }
  return 0;
}


int CTCard::_releaseTerminal(int tid) {
  int err;
  int requestid;

  /* find matching reader */
  err=ChipCard_RequestReleaseReader(&requestid,tid);
  if (err!=CHIPCARD_SUCCESS)
    return err;

  err=_responseLoop(requestid, _timeout);
  if (err!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"No response.");
#endif
    return 1;
  }

  err=ChipCard_CheckReleaseReader(requestid);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"ChipCard error %d",err);
#endif
    return 1;
  }
  if (err!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"Could not release terminal\n");
#endif
    return 1;
  }
  return 0;
}


int CTCard::_statTerminal(unsigned int *status) {
  int err;
  int requestid;
  int result;
  char atr[300];
  int atrlen;

  err=_openTerminal();
  if (err!=CHIPCARD_SUCCESS)
    return err;

  /* find matching reader */
  err=ChipCard_RequestStatReader(&requestid,_terminalHandle);
  if (err!=CHIPCARD_SUCCESS)
    return err;

  err=_responseLoop(requestid, _timeout);
  if (err!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"No response.\n");
#endif
    return 1;
  }

  atrlen=sizeof(atr);
  err=ChipCard_CheckStatReader(requestid, &result, status,
			       atr, &atrlen);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"ChipCard error %d\n",err);
#endif
    return 1;
  }
  if (result!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"Could not stat terminal\n");
#endif
    return 1;
  }
  _atr=string(atr, atrlen);
  return 0;
}


int CTCard::_openTerminal(){
  int err;
  CHIPCARD_READERDESCR *rd;

#if DEBUGMODE>15
  fprintf(stderr,"Opening terminal %08x\n", _terminalId);
#endif
  if (_termOpenCount==0) {
    if (_terminalId==-1) {
      err=_findTerminal(&_terminalId,
			_preferredFlags,
			_preferredFlagsMask);
      if (err)
	return err;
    }

    err=_allocTerminal(_terminalId, &_terminalHandle, &rd);
    if (err)
      return err;
    _readerName=rd->name;
    _readerFlags=rd->flags;
    _readerType=rd->type;
    ChipCard_ReaderDescr_free(rd);
  }
  _termOpenCount++;

  return CHIPCARD_SUCCESS;
}


int CTCard::_openCard(){
  int err;

#if DEBUGMODE>15
  fprintf(stderr,"Opening card\n");
#endif
  if (_openCount==0) {
    err=_openTerminal();
    if (err)
      return err;
    err=_connectTerminal(_terminalHandle);
    if (err) {
      _releaseTerminal(_terminalHandle);
      return err;
    }
  }
  _openCount++;
  return CHIPCARD_SUCCESS;
}


int CTCard::_closeCard(bool force, bool eject){
  int err1, err2;

#if DEBUGMODE>15
  fprintf(stderr,"_closeCard: openCount is %d\n",_openCount);
#endif

  if (_openCount==0)
    return CHIPCARD_SUCCESS;
  _openCount--;
  if (_openCount==0 || force) {
    err1=_disconnectTerminal(_terminalHandle);
    _openCount=0;
    err2=_closeTerminal(true);
    _terminalId=-1;
    if (err1)
      return err1;
    return err2;
  }
  return CHIPCARD_SUCCESS;
}


int CTCard::_closeTerminal(bool force){
  int err;

  err=CHIPCARD_SUCCESS;
  if (_termOpenCount==0)
    return err;
  _termOpenCount--;
  if (_termOpenCount==0 || force) {
    err=_releaseTerminal(_terminalHandle);
    _terminalHandle=-1;
    _termOpenCount=0;
  }
  return err;
}


CTCard::CTCard()
:_terminalId(-1)
,_terminalHandle(-1)
,_preferredFlags(0)
,_preferredFlagsMask(0)
,_openCount(0)
,_termOpenCount(0)
,_cardStatus(0)
,_memsize(0)
,_timeout(30)
,_callback(0)
{
}


CTCard::CTCard(CTCard &c)
{
  *this=c;
  c._openCount=0;
  c._termOpenCount=0;
}


CTCard::~CTCard(){
  if (_openCount) {
#if DEBUGMODE>0
    fprintf(stderr,
	    " LIBCHIPCARD-Warning: \n"
	    " Destroying a CTCard object while it is still open.\n"
	    " Please consider closing a card before destroying it.\n");
#endif
  }
}


CTError CTCard::openCard(){
  int err;

#if DEBUGMODE>15
  fprintf(stderr, "CTCard::openCard()\n");
#endif
  try {
    err=_openCard();
    if (err)
      return CTError("CTCard::openCard()",
		     k_CTERROR_API, err, 0,
		     "Unable to open card");
    return CTError();
  }
  catch (CTError xerr) {
    return CTError("CTCard::openCard()", xerr);
  }
}


CTError CTCard::closeCard(bool force, bool eject){
  int err;

  err=_closeCard(force, eject);
  if (err)
    return CTError("CTCard::closeCard()",
		   k_CTERROR_API, err, 0,
		   "Error closing card");
  return CTError();
}


CTError CTCard::_locateCommandReader(const string &command,
				     const string &reader,
				     string &result) {
  CTError err;
  int pos;
  int tpos;
  string types;
  string ttype;
  string cmd;
  int rv;

  types=cardTypes();
  pos=types.length()-1;

  while(pos>0) {
    tpos=pos;
    ttype.erase();
    while(pos>=0) {
      if (types[pos]==',') {
	if (pos+1<(int)(types.length()))
	  ttype=types.substr(pos+1, tpos-pos);
	if (pos>=0)
	  pos--;
	break;
      }
      pos--;
    }
    if (pos<0)
      ttype=types.substr(0,tpos+1);

    CTMisc::removeBlanks(ttype);

    if (ttype.empty())
      return CTError("CTCard::_locateCommandreader()",
		     k_CTERROR_INVALID,0,0,
		     "Command not found");

    cmd=reader+"/"+ttype+"/"+command;
    rv=ChipCard_ExistsCommand(cmd.c_str());
    if (rv==CHIPCARD_SUCCESS) {
      result=cmd;
      return CTError();
    }
  } // while

  return CTError("CTCard::_locateCommandReader()",
		 k_CTERROR_INVALID,0,0,
		 "Command not found");
}


CTError CTCard::_locateCommand(const string &command,
			       string &result) {
  CTError err;

  // try current reader type first
  err=_locateCommandReader(command, _readerType,result);
  if (err.isOk())
    return CTError();
  // then try "ALL"
  err=_locateCommandReader(command, "all", result);
  if (!err.isOk())
    return CTError("CTCard::_locateCommand",err);
  return CTError();
}



CTError CTCard::doCommand(const string &command,
			  string &cmdCache,
			  string &response,
			  const string &param1,
			  const string &param2,
			  const string &param3,
			  const string &param4,
			  const string &param5){
  int requestid;
  char buffer[300];
  int bufferlen;
  int err;
  int result;
  CTError error;
  int params;

  //fprintf(stderr,"Executing command \"%s\" (%s)\n",
  //        command.c_str(), cmdCache.c_str());

  try {
    params=5;
    if (param5.empty())
      params--;
    if (param4.empty())
      params--;
    if (param3.empty())
      params--;
    if (param2.empty())
      params--;
    if (param1.empty())
      params--;

    if (cmdCache.empty()) {
      error=_locateCommand(command, cmdCache);
      if (!error.isOk()) {
	bufferlen=sizeof(buffer);
	err=ChipCard_LocateCommand(command.c_str(),
				   _readerType.c_str(),
				   cardType().c_str(),
				   buffer,
				   bufferlen);
	if (err!=CHIPCARD_SUCCESS)
	  return CTError("CTCard::doCommand()",
			 k_CTERROR_API, err, 0,
			 command);
	cmdCache=buffer;
      }
    }

    //fprintf(stderr,"Now executing command \"%s\" (%s)\n",
    //        command.c_str(), cmdCache.c_str());

    switch(params) {
    case 0:
      err=ChipCard_RequestCommand2(&requestid,
				   _terminalHandle,
				   cmdCache.c_str(),
				   0);
      break;

    case 1:
      err=ChipCard_RequestCommand2(&requestid,
				   _terminalHandle,
				   cmdCache.c_str(),
				   1, param1.c_str());
      break;

    case 2:
      err=ChipCard_RequestCommand2(&requestid,
				   _terminalHandle,
				   cmdCache.c_str(),
				   2, param1.c_str(), param2.c_str());
      break;

    case 3:
      err=ChipCard_RequestCommand2(&requestid,
				   _terminalHandle,
				   cmdCache.c_str(),
				   3,
				   param1.c_str(),
				   param2.c_str(),
				   param3.c_str());
      break;

    case 4:
      err=ChipCard_RequestCommand2(&requestid,
				   _terminalHandle,
				   cmdCache.c_str(),
				   4,
				   param1.c_str(),
				   param2.c_str(),
				   param3.c_str(),
				   param4.c_str());
      break;

    case 5:
    default:
      err=ChipCard_RequestCommand2(&requestid,
				   _terminalHandle,
				   cmdCache.c_str(),
				   5,
				   param1.c_str(),
				   param2.c_str(),
				   param3.c_str(),
				   param4.c_str(),
				   param5.c_str());
      break;
    } // switch

    if (err!=CHIPCARD_SUCCESS)
      return CTError("CTCard::doCommand()",
		     k_CTERROR_API, err,0,
		     "Error sending command",
		     command);

    err=_responseLoop(requestid, _timeout);
    if (err!=0) {
#if DEBUGMODE>1
      fprintf(stderr,"No response");
#endif
      return CTError("CTCard::doCommand()",
		     k_CTERROR_API, err, 0,
		     "No response");
    }

    bufferlen=sizeof(buffer);
    err=ChipCard_CheckCommand(requestid,
			      &result,
			      buffer,
			      &bufferlen);
    if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
      fprintf(stderr,"Chipcard error %d",err);
#endif
      return CTError("CTCard::doCommand()",
		     k_CTERROR_API, err,0,
		     "Error sending command");
    }
    if (result!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
      fprintf(stderr,"Error in command");
#endif
      return CTError("CTCard::doCommand()",
		     k_CTERROR_API, result, 0,
		     "Error in command");
    }
    if (bufferlen<2)
      return CTError("CTCard::doCommand()",
		     k_CTERROR_API,CHIPCARD_ERROR_BAD_RESPONSE,0,
		     "Error in response");
    error=CTError("CTCard::doCommand()",
		  err,buffer[bufferlen-2],buffer[bufferlen-1]);
    response=string(buffer, bufferlen-2);

#if DEBUGMODE>1
    if (!error.isOk())
      fprintf(stderr,"Result of command: %s: %s\n",command.c_str(),
	      error.errorString().c_str());
#endif
    return error;
  }
  catch (CTError xerr) {
    return CTError("CTCard::doCommand",xerr);
  }
}


CTError CTCard::selectFile(string &fcp, unsigned short fid){
  return doCommand("select_file_w",
		   _cmdSelectFile,
		   fcp,
		   CTMisc::num2string(fid));
}


CTError CTCard::selectFile(string &fcp, const string &fid){
  return doCommand("select_file_id",
		   _cmdSelectFileId,
		   fcp,
		   CTMisc::bin2hex(fid));
}


CTError CTCard::readBinaryRaw(string &data,
			      unsigned short offset,
			      unsigned char size){
  return doCommand("read_binary",
		   _cmdReadBinary,
		   data,
		   CTMisc::num2string(offset),
		   CTMisc::num2string(size));
}


CTError CTCard::updateBinaryRaw(const string &data,
				unsigned short offset){
  string response;

  return doCommand("update_binary",
		   _cmdUpdateBinary,
		   response,
		   CTMisc::num2string(offset),
		   CTMisc::bin2hex(data));
}


CTError CTCard::updateBinary(const string &data,
			     unsigned int pos){
  CTError err;
  unsigned int p;
  string tmp;
  int i,j;

#if DEBUGMODE>15
  fprintf(stderr,"CTCard::updateBinary\n");
#endif
  /* Well, the Towitoko driver seems to have problems writing
   * across 32 bytes boundaries. So we have to check for those
   * boundaries ourselves. At least that driver is able to communicate
   * with I2C memory cards, unlike KOBIL drivers ;-)
   */
  p=0;
  while(p<data.length()) {
    // calculate next boundary at 32 bytes
    j=((pos+p)|CTCARD_BOUNDARY)+1;
    // calculate number of bytes till next 32 byte boundary
    i=j-(pos+p);
    // check number of bytes against end of data
    if (p+i>=data.length())
      i=data.length()-p;
    // now cut out THAT number of bytes
    tmp=data.substr(p,i);
    err=updateBinaryRaw(tmp,pos+p);
    if (!err.isOk())
      return err;
    p+=i;
  }
  return CTError();
}


CTError CTCard::readBinary(string &d,
			   unsigned int pos,
			   unsigned int s){
  CTError err;
  string tmp;
  unsigned int i;

#if DEBUGMODE>15
  fprintf(stderr,
	  "CTCard::readBinary pos=%04x, length=%d\n",
	  pos,s);
#endif

  d.erase();
  while (d.length()<s) {
    tmp.erase();
    i=s-d.length();
    if (i>CTCARD_READSIZE)
      i=CTCARD_READSIZE;
    err=readBinaryRaw(tmp,pos,i);
    d+=tmp;
    pos+=tmp.length();
    if (!err.isOk()) {
      if (!err.isOk(0x6b))
	return err;
#if DEBUGMODE>1
      fprintf(stderr,"LIBCHIPCARD: Stopped reading (readBinary).\n");
#endif
      return CTError();
    } // if error
    if (tmp.empty()){
      return CTError();
    }
  } // while
  return CTError();
}


bool CTCard::isProcessorCard() {
  int err;
  unsigned int status;

  err=_statTerminal(&status);
  if (err!=CHIPCARD_SUCCESS)
    return false;
  return status & CHIPCARD_STATUS_PROCESSOR;
}


bool CTCard::isInserted(){
  int err;
  unsigned int status;

  err=_statTerminal(&status);
  if (err!=CHIPCARD_SUCCESS)
    return false;
  return status & CHIPCARD_STATUS_INSERTED;
}


bool CTCard::isConnected(){
  int err;
  unsigned int status;

  err=_statTerminal(&status);
  if (err!=CHIPCARD_SUCCESS)
    return false;
  return status & CHIPCARD_STATUS_CONNECTED;
}


string CTCard::cardType(){
  return "CTCard";
}


string CTCard::cardTypes(){
  return "CTCard";
}


CTError CTCard::reopenCard(){
#if DEBUGMODE>15
  fprintf(stderr,"CTCard::reopenCard\n");
#endif

#if DEBUGMODE>0
  fprintf(stderr,"LIBCHIPCARD: The method \"CTCard::reopenCard()\"\n"
	  "has been called. Since THIS class does nothing on this method\n"
	  "I guess you have an error in your program.\n");
#endif
  return CTError("CTCard::reopenCard()",
		 k_CTERROR_INVALID,
		 0,
		 0,
		 "CTCard::reopenCard() should be overloaded.");
}


CTError CTCard::sendAPDU(const string &apdu, string &response) {
  int err;
  int requestid;
  char buffer[300];
  int bufferlen;
  int result;
  CTError error;

  err=ChipCard_RequestCommand(&requestid,
			      _terminalHandle,
			      apdu.c_str(),
			      apdu.length());
  err=_responseLoop(requestid, _timeout);
  if (err!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"No response");
#endif
    return CTError("CTCard::doCommand()",
		   k_CTERROR_API, err, 0,
		   "No response");
  }

  bufferlen=sizeof(buffer);
  err=ChipCard_CheckCommand(requestid,
			    &result,
			    buffer,
			    &bufferlen);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"Chipcard error %d",err);
#endif
    return CTError("CTCard::doCommand()",
		   k_CTERROR_API, err, 0,
		   "Error sending command");
  }
  if (result!=0) {
#if DEBUGMODE>1
    fprintf(stderr,"Error in command");
#endif
    return CTError("CTCard::doCommand()",
		   err,0,0,
		   "Error in command");
  }
  response.assign(buffer, bufferlen);
  return CTError();
}



CTError CTCard::execCommand(CTCommand &cmd) {
  CTError error;
  string response;

  error=sendAPDU(cmd.toString(), response);
  if (!error.isOk())
    return CTError("CTCard::execCommand", error);

  if (response.length()<2)
    return CTError("CTCard::doCommand()",
		   k_CTERROR_API, CHIPCARD_ERROR_BAD_RESPONSE,0,
		   "Error in response");
  error=CTError("CTCard::execCommand()",
		0,
		response[response.length()-2],
		response[response.length()-1]);
  cmd.setData(response.substr(0,response.length()-2));
  cmd.setSw1(response[response.length()-2]);
  cmd.setSw2(response[response.length()-1]);

  return error;
}


int CTCard::startWaitForCard(int &requestid,
			     bool next,
			     unsigned int readerFlags,
			     unsigned int readerFlagsMask,
			     unsigned int status,
			     unsigned int statusMask){
  int err;

  err=_openLibrary();
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"Chipcard error %d\n",err);
#endif
    return err;
  }

  /* find matching reader */
  err=ChipCard_RequestWaitReader(&requestid,
				 next,
				 "",
				 readerFlags,
				 readerFlagsMask,
				 status, statusMask);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"Chipcard error %d\n",err);
#endif
    return err;
  }

  return CHIPCARD_SUCCESS;
}


bool CTCard::waitForCard(int timeout,
			 bool next,
			 unsigned int &readerFlags,
			 unsigned int readerFlagsMask,
			 unsigned int &status,
			 unsigned int statusMask){
  int err;
  int err2;
  int requestid;

  if (_openCount>0)
    return true;

  err=startWaitForCard(requestid,
		       next,
		       readerFlags,
		       readerFlagsMask,
		       status,
		       statusMask);
  if (err!=CHIPCARD_SUCCESS)
    return false;

  err=checkWaitForCard(requestid,
		       timeout,
		       readerFlags,
		       status);
  err2=stopWaitForCard(requestid);
  if (err!=CHIPCARD_SUCCESS)
    return false;

  if (err2!=CHIPCARD_SUCCESS)
    return false;

  return true;
}


int CTCard::checkWaitForCard(int requestid,
			     int timeout,
			     unsigned int &readerFlags,
			     unsigned int &status){
  int ltid;
  unsigned int lstatus;
  unsigned int lflags;
  int err;

  err=_responseLoop(requestid, timeout);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"Chipcard error %d\n",err);
#endif
    return err;
  }

  err=ChipCard_CheckWaitReader(requestid,
			       &ltid,
			       &lstatus,
			       &lflags);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"Chipcard error %d\n",err);
#endif
    return err;
  }
  readerFlags=lflags;
  status=lstatus;
  _terminalId=ltid;
  return CHIPCARD_SUCCESS;
}


int CTCard::stopWaitForCard(int requestid){
  int err;

  err=ChipCard_StopWaitReader(requestid);
  if (err!=CHIPCARD_SUCCESS) {
#if DEBUGMODE>1
    fprintf(stderr,"Chipcard error %d\n",err);
#endif
    return err;
  }
  return CHIPCARD_SUCCESS;
}





void CTCard::setCallBack(CallBack cb, void *p){
  _callback=cb;
  _callbackParam=p;
}


bool CTCard::findReader(unsigned int readerFlags,
			unsigned int readerFlagsMask){
  int err;
  int tid;

  if (_openCount>0)
    return false;
  err=_findTerminal(&tid,
		    readerFlags,
		    readerFlagsMask);
  if (err!=CHIPCARD_SUCCESS) {
    DBG_ERROR("Error: %d\n",err);
    return false;
  }
#if DEBUGMODE>1
  fprintf(stderr,"findReader: Selecting reader %08x\n",_terminalId);
#endif
  _terminalId=tid;
  return true;
}


CTError CTCard::getCard(int timeout,
			bool next,
			unsigned int readerFlags,
			unsigned int readerFlagsMask,
			unsigned int status,
			unsigned int statusMask) {

  int requestid;
  unsigned int lflags;
  unsigned int lstatus;
  int rv;
  CTError err;
  time_t startTime;

  // timouts > 1h are not tolerated
  if (timeout>3600)
    return CTError("CTCard::getCard()",
		   k_CTERROR_INVALID, 0,0,
		   "timout too long, please consider using value 0");

  // set starting time
  startTime=time(NULL);
  rv=startWaitForCard(requestid,
		      next,
		      readerFlags,
		      readerFlagsMask,
		      status,
		      statusMask);
  if (rv!=CHIPCARD_SUCCESS) {
    return CTError("CTCard::getCard()",
		   k_CTERROR_API, rv,0,
		   "startWaitForCard");
  }

  while(1) {
    int localtimeout;
    int diff;

    if (timeout) {
      // calculate local timeout (remaining time for this loop)
      diff=(int)difftime(time(NULL),startTime);
      if (diff>=timeout) {
	// timeout reached, stop waiting and return with error
	stopWaitForCard(requestid);
	return CTError("CTCard::getCard()",
		       k_CTERROR_API, CHIPCARD_ERROR_NO_CARD, 0,
		       "timed out");
      }
      localtimeout=timeout-diff;
    }
    else
      localtimeout=0;

    // actually wait for a card
    rv=checkWaitForCard(requestid,
			timeout,
			lflags,
			lstatus);
    if (rv!=CHIPCARD_SUCCESS) {
      // no card within timeout or user aborted
      stopWaitForCard(requestid);
      return CTError("CTCard::getCard()",
		     k_CTERROR_API, rv,0,
		     "checkWaitForCard");
    }

    // try to open the card
    err=openCard();
    if (err.isOk()) {
      // card is open, check it
      err=checkCard(lflags,
		    lstatus);
      if (err.isOk()) {
	// card matches, stop waiting
	rv=stopWaitForCard(requestid);
	if (rv!=CHIPCARD_SUCCESS) {
	  return CTError("CTCard::getCard()",
			 k_CTERROR_API, rv,0,
			 "stopWaitForCard");
	}
	// finished
	return CTError();
      }
      else {
	// card does not match, close it and wait for next
	err=closeCard();
	if (!err.isOk()) {
	  fprintf(stderr,"Warning: %s\n",err.errorString().c_str());
          // check if aborted (e.g. by user)
	  if (err.code()==k_CTERROR_API &&
	      err.subcode1()==CHIPCARD_ERROR_ABORTED)
	    return CTError("CTCard::getCard()",err);
	}
      }
    }
  } // while
}


CTError CTCard::checkCard(unsigned int readerFlags,
			  unsigned int status){
  return CTError();
}





