/***************************************************************************
 $RCSfile: inetaddr.c,v $
                             -------------------
    cvs         : $Id: inetaddr.c,v 1.4 2003/01/11 12:58:45 aquamaniac Exp $
    begin       : Tue Oct 02 2002
    copyright   : (C) 2002 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef __declspec
# if BUILDING_CHIPCARD_DLL
#  define CHIPCARD_API __declspec (dllexport)
# else /* Not BUILDING_CHIPCARD_DLL */
#  define CHIPCARD_API __declspec (dllimport)
# endif /* Not BUILDING_CHIPCARD_DLL */
#else
# define CHIPCARD_API
#endif


#include <stdlib.h>
#include <assert.h>
#include "chameleon/inetaddr.h"
#include <string.h>


// forward declaration
const char *InetAddr_ErrorString(int c);


int inetaddr_is_initialized=0;
ERRORTYPEREGISTRATIONFORM inetaddr_error_descr= {
    InetAddr_ErrorString,
    0,
    "InetAddr"};



ERRORCODE InetAddr_ModuleInit(){
    if (!inetaddr_is_initialized) {
	if (!Error_RegisterType(&inetaddr_error_descr))
	    return Error_New(0,
			     ERROR_SEVERITY_ERR,
			     ERROR_TYPE_ERROR,
			     ERROR_COULD_NOT_REGISTER);
	inetaddr_is_initialized=1;
    }
    return 0;
}


ERRORCODE InetAddr_ModuleFini(){
    if (inetaddr_is_initialized) {
	inetaddr_is_initialized=0;
	if (!Error_UnregisterType(&inetaddr_error_descr))
	    return Error_New(0,
			     ERROR_SEVERITY_ERR,
			     ERROR_TYPE_ERROR,
			     ERROR_COULD_NOT_UNREGISTER);
    }
    return 0;
}


const char *InetAddr_ErrorString(int c) {
    const char *s;

    switch(c) {
    case INETADDR_ERROR_MEMORY_FULL:
	s="Memory full";
	break;
    case INETADDR_ERROR_BAD_ADDRESS:
	s="Bad address";
	break;
    case INETADDR_ERROR_BUFFER_OVERFLOW:
	s="Buffer overflow";
	break;
    case INETADDR_ERROR_HOST_NOT_FOUND:
	s="Host not found";
	break;
    case INETADDR_ERROR_NO_ADDRESS:
	s="No address";
	break;
    case INETADDR_ERROR_NO_RECOVERY:
	s="No recovery";
	break;
    case INETADDR_ERROR_TRY_AGAIN:
	s="Try again";
	break;
    case INETADDR_ERROR_UNKNOWN_DNS_ERROR:
	s="Unknown DNS error";
	break;
    default:
	s=(const char*)0;
    } // switch
    return s;
}


ERRORCODE InetAddr_Create(INETADDRESSPTR ia, const char *addr, int port){
    ERRORCODE err;

    assert(ia);
    // init data with defaults
    ia->address.sin_port=htons(0);
    ia->address.sin_family=PF_INET;

    // set address and port
    err=InetAddr_SetAddress(ia,addr);
    if (!Error_IsOk(err))
	err=InetAddr_SetName(ia,addr);
    if (Error_IsOk(err))
	err=InetAddr_SetPort(ia,port);
    if (!Error_IsOk(err)) {
	return err;
    }
    return 0;
}


ERRORCODE InetAddr_Destroy(INETADDRESSPTR ia) {
    assert(ia);
    return 0;
}


ERRORCODE InetAddr_SetAddress(INETADDRESSPTR ia, const char *addr){
    assert(ia);

    // reset
    ia->address.sin_family=PF_INET;
    ia->address.sin_addr.s_addr=0;

    if (addr) {
	// ok, address to be set
	ia->address.sin_addr.s_addr=inet_addr(addr);
	if (ia->address.sin_addr.s_addr==INADDR_NONE)
	    // bad address, so maybe it rather is a name
	    return Error_New(0,
			     ERROR_SEVERITY_ERR,
			     inetaddr_error_descr.typ,
			     INETADDR_ERROR_BAD_ADDRESS);
    }
    return 0;
}


// internal function
int InetAddr_TranslateHError(int herr) {
    int rv;

    switch(herr) {
    case WSAHOST_NOT_FOUND:
	rv=INETADDR_ERROR_HOST_NOT_FOUND;
	break;
    case WSANO_ADDRESS:
	rv=INETADDR_ERROR_NO_ADDRESS;
	break;
    case WSANO_RECOVERY:
	rv=INETADDR_ERROR_NO_RECOVERY;
	break;
    case WSATRY_AGAIN:
	rv=INETADDR_ERROR_TRY_AGAIN;
	break;
    default:
	rv=INETADDR_ERROR_UNKNOWN_DNS_ERROR;
	break;
    } // switch
    return rv;
}


ERRORCODE InetAddr_SetName(INETADDRESSPTR ia, const char *name){
    struct hostent *he;

    assert(ia);
    // try to resolve name
    he=gethostbyname(name);
    if (!he)
	return Error_New(0,
			 ERROR_SEVERITY_ERR,
			 inetaddr_error_descr.typ,
			 InetAddr_TranslateHError(h_errno));
    // name resolved, store address
    memcpy(&(ia->address.sin_addr),he->h_addr_list[0],sizeof(struct in_addr));
    return 0;
}


ERRORCODE InetAddr_GetAddress(const INETADDRESS *ia, char *buffer, int bsize){
    const char *s;

    assert(ia);
    assert(buffer);
    s=inet_ntoa(ia->address.sin_addr);
    assert(s);
    if (strlen(s)+1>bsize)
	return Error_New(0,
			 ERROR_SEVERITY_ERR,
			 inetaddr_error_descr.typ,
			 INETADDR_ERROR_BUFFER_OVERFLOW);
    strcpy(buffer,s);
    return 0;
}


ERRORCODE InetAddr_GetName(const INETADDRESS *ia, char *buffer, int bsize){
    struct hostent *he;
    struct in_addr lia;

    assert(ia);

    // resolve name from address
    lia=ia->address.sin_addr;
    he=gethostbyaddr((char*)&lia,sizeof(lia),PF_INET);
    if (!he)
	return Error_New(0,
			 ERROR_SEVERITY_ERR,
			 inetaddr_error_descr.typ,
			 InetAddr_TranslateHError(h_errno));

    // copy name into given buffer
    assert(he->h_name);
    if (strlen(he->h_name)+1>bsize)
	return Error_New(0,
			 ERROR_SEVERITY_ERR,
			 inetaddr_error_descr.typ,
			 INETADDR_ERROR_BUFFER_OVERFLOW);
    // copy the name into the buffer
    strcpy(buffer,he->h_name);

    // finished
    return 0;
}


int InetAddr_GetPort(const INETADDRESS *ia){
    assert(ia);

    return ntohs(ia->address.sin_port);
}


ERRORCODE InetAddr_SetPort(INETADDRESSPTR ia, int port){
    assert(ia);

    ia->address.sin_port=htons(port);
    return 0;
}






